use std::{
    collections::BTreeMap,
    future::{self, Future},
    ops::Deref,
    path::{Path, PathBuf},
    pin::Pin,
    sync::Arc,
};

use rust_builder::RustActionsBuilder;
use serde::{Deserialize, Serialize};

use crate::state::validated_project::Value;

pub mod builder;

pub mod rust_builder {
    use std::{
        env::temp_dir,
        path::{Path, PathBuf},
    };

    use serde::Deserialize;

    use crate::actions::CuddleActionsSchema;

    use super::ExecutableActions;

    pub struct RustActionsBuilder {
        root_path: PathBuf,
    }

    impl RustActionsBuilder {
        pub fn new(root_path: PathBuf) -> Self {
            Self { root_path }
        }

        fn actions_path(&self) -> Option<PathBuf> {
            let actions_path = self.root_path.join("actions/rust");
            if !actions_path.exists() {
                return None;
            }

            Some(actions_path)
        }

        fn global_registry(&self) -> anyhow::Result<Option<PathBuf>> {
            if let Some(dir) = dirs::cache_dir().map(|c| c.join("sh.cuddle/registry/actions/rust"))
            {
                if !dir.exists() {
                    std::fs::create_dir_all(&dir)?;
                }

                Ok(Some(dir))
            } else {
                Ok(None)
            }
        }

        pub async fn build(&self) -> anyhow::Result<ExecutableActions> {
            tracing::debug!("building rust action: {}", self.root_path.display());

            let Some(path) = self.actions_path() else {
                anyhow::bail!(
                    "action was not found: {}",
                    self.root_path.display().to_string()
                );
            };

            let actions_registry = self
                .global_registry()?
                .ok_or(anyhow::anyhow!("failed to find global registry"))?;

            let staging_id = uuid::Uuid::new_v4();
            let actions_temp = temp_dir()
                .join("cuddle/actions")
                .join(staging_id.to_string());
            std::fs::create_dir_all(&actions_temp)?;

            let mut hasher = blake3::Hasher::new();

            tracing::debug!("moving file into: {}", actions_temp.display());
            for entry in walkdir::WalkDir::new(&path) {
                let entry = entry?;
                let full_path = entry.path();
                let rel_path = full_path
                    .strip_prefix(path.canonicalize()?.to_string_lossy().to_string())?
                    .to_string_lossy();

                if rel_path.contains("target/")
                    || rel_path.contains(".cuddle/")
                    || rel_path.contains(".git/")
                {
                    continue;
                }

                let metadata = entry.metadata()?;
                if metadata.is_file() {
                    let temp_file_path = actions_temp.join(rel_path.to_string());

                    if let Some(temp_parent) = temp_file_path.parent() {
                        std::fs::create_dir_all(temp_parent)?;
                    }

                    std::fs::copy(entry.path(), temp_file_path)?;

                    hasher.update(entry.path().to_string_lossy().as_bytes());
                    let file_bytes = tokio::fs::read(entry.path()).await?;
                    hasher.update(&file_bytes);
                }
            }
            let digest = hasher.finalize().to_hex().to_string();

            let action_index = actions_registry.join(digest);
            if action_index.exists() {
                tracing::debug!("action already exists in: {}", action_index.display());

                return self.get_actions(&action_index.join("action")).await;
            }
            std::fs::create_dir_all(&action_index)?;

            tracing::debug!("building rust code: {}", actions_temp.display());
            let mut cmd = tokio::process::Command::new("cargo");
            let output = cmd
                .args(vec!["build", "--release"])
                .current_dir(&actions_temp)
                .output()
                .await?;
            if !output.status.success() {
                anyhow::bail!(
                    "cargo build failed: {}",
                    std::str::from_utf8(&output.stderr)?
                );
            }

            let temp_file_bin_path = actions_temp.join("target/release/action");
            tokio::fs::copy(temp_file_bin_path, action_index.join("action")).await?;

            self.get_actions(&action_index.join("action")).await
        }

        pub async fn get_actions(&self, action_path: &Path) -> anyhow::Result<ExecutableActions> {
            tracing::debug!("querying schema: {}", action_path.display());
            let output = tokio::process::Command::new(action_path.to_string_lossy().to_string())
                .arg("schema")
                .output()
                .await?;

            let actions: CuddleActionsSchema = match serde_json::from_slice(&output.stdout) {
                Ok(output) => output,
                Err(e) => {
                    let schema_output = std::str::from_utf8(&output.stdout)?;

                    anyhow::bail!(
                        "failed to query schema: {} {}",
                        e.to_string(),
                        schema_output
                    )
                }
            };

            Ok(actions.to_executable(action_path)?)
        }
    }
}

pub struct Actions {
    variants: Vec<ActionVariant>,
}

impl Actions {
    pub async fn new(path: &Path, value: &Value) -> anyhow::Result<Option<Self>> {
        let Some(project) = value.get(&["project", "actions"]) else {
            tracing::debug!("found no actions folder");
            return Ok(None);
        };

        let mut variants = Vec::default();
        if let Some(Value::Bool(true)) = project.get(&["rust"]) {
            tracing::debug!("rust actions enabled");
            variants.push(ActionVariant::Rust {
                root_path: path.to_path_buf(),
            });
        } else {
            tracing::trace!("rust actions not enabled");
        }

        Ok(Some(Self { variants }))
    }

    pub async fn build(&mut self) -> anyhow::Result<ExecutableActions> {
        let mut executable_actions = Vec::default();

        for variant in &mut self.variants {
            match variant {
                ActionVariant::Rust { root_path } => {
                    let actions = RustActionsBuilder::new(root_path.clone()).build().await?;

                    executable_actions.push(actions);
                }
                ActionVariant::Docker => todo!(),
            }
        }

        let mut exec_actions = ExecutableActions::default();
        for mut actions in executable_actions {
            exec_actions.actions.append(&mut actions.actions);
        }

        Ok(exec_actions)
    }
}

#[derive(Debug, Deserialize)]
struct CuddleActionsSchema {
    actions: Vec<CuddleActionSchema>,
}

#[derive(Debug, Deserialize)]
struct CuddleActionSchema {
    name: String,
}

pub enum ActionVariant {
    Rust { root_path: PathBuf },
    Docker,
}

#[derive(Default)]
pub struct ExecutableActions {
    pub actions: Vec<ExecutableAction>,
}

pub struct ExecutableAction {
    pub name: String,
    pub description: String,
    pub flags: BTreeMap<String, ExecutableActionFlag>,
    call_fn: LazyResolve,
}

impl ExecutableAction {
    pub async fn call(&self) -> anyhow::Result<()> {
        // Bad hack until .call becomes stable
        (self.call_fn.0)().await?;

        Ok(())
    }
}

#[derive(Debug, Serialize, Deserialize)]
pub enum ExecutableActionFlag {
    String,
    Bool,
}

impl CuddleActionsSchema {
    fn to_executable(self, action_path: &Path) -> anyhow::Result<ExecutableActions> {
        Ok(ExecutableActions {
            actions: self
                .actions
                .into_iter()
                .map(|a| {
                    let name = a.name.clone();
                    let action_path = action_path.to_string_lossy().to_string();

                    ExecutableAction {
                        name: a.name,
                        description: String::new(),
                        flags: BTreeMap::default(),
                        call_fn: LazyResolve::new(Box::new(move || {
                            let name = name.clone();
                            let action_path = action_path.clone();

                            Box::pin(async move {
                                tracing::debug!("calling: {}", name);
                                let mut cmd = tokio::process::Command::new(action_path);
                                cmd.args(["do", &name]);

                                let output = cmd.output().await?;
                                let stdout = std::str::from_utf8(&output.stdout)?;
                                for line in stdout.lines() {
                                    println!("{}: {}", &name, line);
                                }

                                tracing::debug!("finished call for output: {}", &name);

                                Ok(())
                            })
                        })),
                    }
                })
                .collect(),
        })
    }
}

struct LazyResolve(
    Arc<dyn Fn() -> Pin<Box<dyn Future<Output = anyhow::Result<()>> + Send>> + Send + Sync>,
);

impl LazyResolve {
    pub fn new(
        func: Box<
            dyn Fn() -> Pin<Box<dyn Future<Output = anyhow::Result<()>> + Send>> + Send + Sync,
        >,
    ) -> Self {
        Self(Arc::new(func))
    }
}

impl Deref for LazyResolve {
    type Target =
        Arc<dyn Fn() -> Pin<Box<dyn Future<Output = anyhow::Result<()>> + Send>> + Send + Sync>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}
